package com.jpmorrsn.javaFBP;

import java.util.*;

/**
 * All verbs must extend this class, defining its three abstract methods
 * <code>openPorts<code>, <code>introspect<code>, and <code>execute</code>.
 *
 * There will be an instance of this class for every node in the network.
 */

public abstract class Component extends Thread {

	static final String copyright = "Copyright 1999, 2000, 2001, 2002, J. Paul Morrison.  At your option, you may copy, "
			+ "distribute, or make derivative works under the terms of the Clarified Artistic License, "
			+ "based on the Everything Development Company's Artistic License.  A document describing "
			+ "this License may be found at http://www.jpaulmorrison.com/fbp/artistic2.htm. "
			+ "THERE IS NO WARRANTY; USE THIS PRODUCT AT YOUR OWN RISK.";

	// All the input ports are stored here, keyed by name.
	HashMap<String, InputPort> inputPorts = new HashMap<String, InputPort>();

	// All the output ports are stored here, keyed by name.
	HashMap<String, OutputPort> outputPorts = new HashMap<String, OutputPort>();

	/**
	 * This is a stack which is made available to each component. Rather than
	 * declaring a special API, we make it <code>protected</code> so that the
	 * component can use the regular java.util.Stack API.
	 */

	protected Stack stack = new Stack();

	// This is the automatic input port named "*IN"
	InputPort autoInput;

	// This is the automatic output port named "*OUT"
	OutputPort autoOutput;

	// This is a count of packets owned by this component.
	// Whenever the component deactivates, the count must be zero.
	int packetCount = 0;

	// This field maintains the Component status

	public enum StatusValues {
		NOT_STARTED, ACTIVE, DORMANT, SUSP_RECV, SUSP_SEND, TERMINATED;
		private int value;

		public int getValue() {
			return value;
		}
	}

	StatusValues status;

	public Network network;

	boolean terminating;

	boolean mustrun = false;

	boolean selfStarting = false;

	long packetNumber = 0; // debugging only

	// HashMap <String,Packet> ownedPackets = null; //debugging only
	public Network mother = null; // added for subnet support

	public Component() {
		super();
	}

	/**
	 * This method is called from other parts of the system to activate this
	 * Component if it needs to be. This will start its thread if needed, and if
	 * already started, will notify() it.
	 */
	synchronized void activate() {

		if (isTerminated())
			return;
		if (!isAlive())
			start();
		else                // was 
			notifyAll();    //missing 

	}

	/**
	 * This method creates a new Packet using a type and String, say for
	 * brackets
	 * 
	 * @return javaflow.Packet
	 * @param s
	 *            java.lang.String
	 */
	public Packet create(int newType, String s) {
		return new Packet(newType, s, this);
	}

	/**
	 * This method creates a Packet pointing at an Object
	 * 
	 * @return javaflow.Packet
	 * @param o
	 *            java.lang.Object
	 */
	public Packet create(Object o) {
		return new Packet(o, this);
	}

	/**
	 * This method is invoked for each activation of the Component. The
	 * component is activated when data arrives at an input port, and executes
	 * until it returns. If all upstream components have closed down, it can
	 * close down; otherwise it is reactivated when more data arrives.
	 */
	void doActivation() throws Throwable {
		packetCount = 0;
		if (autoInput != null) {
			Packet p = autoInput.receive();
			p.drop();
		}

		terminating = true;

		Iterator allin = (inputPorts.values()).iterator();
		while (allin.hasNext()) {

			InputPort port = (InputPort) allin.next();
			if (port instanceof InitializationConnection) {
				InitializationConnection ic = (InitializationConnection) port;
				ic.packet = create(ic.content);
			} else if (port instanceof Connection) {
				Connection cnxt = (Connection) port;
				if (!cnxt.isDrained())
					terminating = false;
			}

		}

		network.GenTraceLine("Activated: " + getName());

		try {
			execute(); // do one activation!
		} catch (Throwable t) {
			FlowError.complain(t.getMessage() + " - component: "
					+ this.getName());
		}
		;
		network.GenTraceLine("Deactivated: " + getName());

		if (packetCount != 0) {
			network.GenTraceLine(this.getName() + " deactivated holding "
					+ packetCount + " packets");
			FlowError
					.complain(packetCount
							+ " packets not disposed of during component deactivation of "
							+ this.getName());
		}

		boolean allClosed = true;
		boolean hasData = false;

		allin = (inputPorts.values()).iterator();
		while (allin.hasNext()) {
			InputPort port = (InputPort) (allin.next());
			if (port == autoInput)
				continue;
			if (!(port instanceof Connection))
				continue;
			Connection cnxt = (Connection) port;
			if (cnxt.count() > 0) {
				allClosed = false;
				hasData = true;
				break;
			}
			if (!(cnxt.isClosed())) {
				allClosed = false;
				break;
			}
		}
		if (allClosed) {

			if (autoOutput != null) {
				Packet p = create(null);
				if (!(autoOutput.send(p)))
					p.drop();
			}

			terminating = true;

			return;
		}

		if (!hasData) {

			status = StatusValues.DORMANT;
			network.GenTraceLine("Dormant: " + getName());
			synchronized (this) {
				try {
					wait();
				} catch (IllegalMonitorStateException e) {
					FlowError.complain("Illegal monitor state");
				} catch (InterruptedException e) {
				}
				;
			}
			status = StatusValues.ACTIVE;
			network.GenTraceLine("Active: " + getName());
		}

	}

	/**
	 * Verbs (components) override this method with their program. This method
	 * is called from a private thread.
	 * 
	 * @exception Throwable
	 *                if there is any error or exception in the program; this is
	 *                a kludge until we work out how to integrate exceptions
	 *                properly.
	 */

	protected abstract void execute() throws Throwable;

	public HashMap<String, InputPort> getInports() {
		return inputPorts;
	}

	public HashMap<String, OutputPort> getOutports() {
		return outputPorts;
	}

	public int getPacketCount() {
		return packetCount;
	}

	/**
	 * Obtain status of component
	 * 
	 * @return int
	 */
	public StatusValues getStatus() {
		return status;
	}

	/**
	 * Verbs override this method to provide information about their ports to a
	 * <cite>Flowbox</cite>. Although the result is declared as an array of
	 * <code>Object</code>s, it must actually contain a <code>String</code>
	 * describing the verb as a whole, followed by a sequence of <i>descriptor</i>s
	 * with one per port, where a <i>descriptor</i> is represented by four
	 * <code>Object</code>s in a row. These are: the port name (as a
	 * <code>String</code>); the port type (one of <code>input</code>,
	 * <code>input[]</code>, <code>output</code>, <code>output[]</code>,
	 * or <code>parameter</code>); the Java type of objects passing through
	 * the port (either a <code>Class</code> or <code>null</code>, which
	 * means the same as <code>Object.class</code>); a human-readable
	 * description (a <code>String</code>) suitable for use in the sentence
	 * <cite>"This port contains ..."</cite>.
	 */

	protected abstract Object[] introspect();

	/**
	 * This method returns <code> true </code> if the component has terminated
	 * 
	 * @return boolean
	 */
	public boolean isTerminated() {
		return status == StatusValues.TERMINATED;
	}

	/**
	 * Verbs call this method from their <code>openPorts</code> method to open
	 * an InputPort, either a regular port or a parameter port.
	 * 
	 * @param name
	 *            the name of the InputPort
	 * @return the InputPort, which should be assigned to an instance variable
	 *         of the verb
	 */

	protected final InputPort openInput(String name) {
		if (name.startsWith("*"))
			FlowError.complain("Attempt to open * port: " + this.getName()
					+ ' ' + name);
		InputPort port = inputPorts.get(name);
		if (port == null)
			return NullConnection.singleton;
		else
			return port;
	}

	/**
	 * Verbs call this method from their <code>openPorts</code> method to open
	 * an array of InputPorts. The size of the Array is determined by the
	 * network to which this component belongs, and thus the verb does not know
	 * it until <code>execute</code> time.
	 * 
	 * @param the
	 *            name of the InputPort array
	 * @return the InputPort[] object, which should be assigned to an instance
	 *         variable.
	 */

	// Array ports are stored individually in the inputPorts
	// and outputPorts variables with names like *NAME*X, where
	// X is a single character encoding the index number.
	// We need to run through inputPorts twice to find out
	// the largest index number (but not less than zero)
	// so we can create a fixed-size Java array to return.
	protected final InputPort[] openInputArray(String name) {
		int highIndex = -1;
		String match = "*" + name + "*";

		Iterator allkeys = (inputPorts.keySet()).iterator();
		while (allkeys.hasNext()) {
			String key = (String) (allkeys.next());
			if (key.startsWith(match)) {
				int index = key.charAt(key.length() - 1);
				if (index > highIndex)
					highIndex = index;
			}
		}
		if (highIndex == -1)
			FlowError.complain("No elements defined in input array: "
					+ this.getName() + ' ' + name);
		InputPort[] array = new InputPort[highIndex + 1];

		for (int i = 0; i < array.length; i++)
			array[i] = null; // NullConnection.singleton;

		allkeys = (inputPorts.keySet()).iterator();
		while (allkeys.hasNext()) {
			String key = (String) (allkeys.next());
			if (key.startsWith(match)) {
				int index = key.charAt(key.length() - 1);
				array[index] = (inputPorts.get(key));

			}
		}

		return array;
	}

	/**
	 * Verbs call this method from their <code>openPorts</code> method to open
	 * an output port.
	 * 
	 * @param name
	 *            the name of the OutputPort
	 * @return the OutputPort, which should be assigned to an instance variable
	 */

	protected final OutputPort openOutput(String name) {
		if (name.startsWith("*"))
			FlowError.complain("Attempt to open * port");
		OutputPort port = outputPorts.get(name);
		if (port == null)
			FlowError.complain("Output port not found: " + this.getName() + ' '
					+ name);

		return port;
	}

	/**
	 * Verbs call this method from their <code>openPorts</code> method to open
	 * an array of OutputPorts. The size of the Array is determined by the
	 * network to which this component belongs, and thus the verb does not know
	 * it until <code>execute</code> time.
	 * 
	 * @param the
	 *            name of the OutputPort array
	 * @return the OutputPort[] object, which should be assigned to an instance
	 *         variable.
	 */

	// See openInputArray.
	protected final OutputPort[] openOutputArray(String name) {
		int highIndex = -1;
		String match = "*" + name + "*";

		Iterator allkeys = (outputPorts.keySet()).iterator();
		while (allkeys.hasNext()) {
			String key = (String) (allkeys.next());
			if (key.startsWith(match)) {
				int index = key.charAt(key.length() - 1);
				if (index > highIndex)
					highIndex = index;
			}
		}
		if (highIndex == -1)
			FlowError.complain("No elements defined in output array :"
					+ this.getName() + ' ' + name);
		OutputPort[] array = new OutputPort[highIndex + 1];

		for (int i = 0; i < array.length; i++)
			array[i] = null;

		allkeys = (outputPorts.keySet()).iterator();
		while (allkeys.hasNext()) {
			String key = (String) (allkeys.next());
			if (key.startsWith(match)) {
				int index = key.charAt(key.length() - 1);
				array[index] = outputPorts.get(key);

			}
		}

		return array;
	}

	/**
	 * Verbs override this method to open their ports. This method is called
	 * from the network's main thread, and should only call
	 * <code>openInput</code>, <code>openInputArray</code>,
	 * <code>openOutput</code>,
	 * <code>openOutputArray<code>, <code>InputPort.setType</code>,
	 * and <code>OutputPort.setType</code>.
	 */

	protected abstract void openPorts();

	/**
	 * This method must be <code>public</code> because <code>Component</code>
	 * inherits from <code>Thread</code>, but should never be called except
	 * by the JVM.
	 * 
	 * If this method is started in 'terminating' mode, the code is not
	 * executed, but upstream and downstream ports are still closed.
	 */

	public final void run() {

		if (this instanceof MustRun)
			mustrun = true;

		if (!terminating || mustrun) {

			status = StatusValues.ACTIVE;
			network.GenTraceLine("Started: " + getName());

			autoInput = inputPorts.get("*IN"); // ?? right place?
			autoOutput = outputPorts.get("*OUT");

			while ((!terminating || mustrun) && !isTerminated()) {

				try {
					doActivation();
				}
				// catch (Exception ex) {} // prevent duplicate logging
				// catch (InterruptedException ex) {
				// System.out.println("Component closing down due to interrupt:
				// " +
				// getName());
				// status = TERMINATED;
				// }
				catch (Throwable t) {
					if (!(t instanceof ThreadDeath)) {
						// status = StatusValues.TERMINATED;
						// mother.notify();
						System.out.println("Exception:" + t.getMessage());
						System.out.println(t.getClass());
						t.printStackTrace();
					}
					mother.indicateTerminated(this);

				}
				mustrun = false;

				if (selfStarting)
					break;
			}

			// network.GenTraceLine("Terminated: " + getName());

		}

		Iterator allin = (inputPorts.values()).iterator();
		while (allin.hasNext()) {

			InputPort ip = (InputPort) allin.next();
			ip.close();

		}

		Iterator allout = (outputPorts.values()).iterator();

		while (allout.hasNext()) {

			OutputPort op = (OutputPort) allout.next();
			op.close();

		}
		mother.indicateTerminated(this);
	}
}
