package com.jpmorrsn.javaFBP;

import java.util.*;

import com.jpmorrsn.javaFBP.Component.StatusValues;

/**
 * This class implements buffering between Component threads. One is created
 * behind the scenes whenever two ports are connected. This class was founded on
 * Doug Lea's BoundedBufferVST class from his book _Concurrent Programming in
 * Java_, page 100.
 */

final class Connection implements InputPort {

	static final String copyright = "Copyright 1999, 2000, 2001, 2002, J. Paul Morrison.  At your option, you may copy, "
			+ "distribute, or make derivative works under the terms of the Clarified Artistic License, "
			+ "based on the Everything Development Company's Artistic License.  A document describing "
			+ "this License may be found at http://www.jpaulmorrison.com/fbp/artistic2.htm. "
			+ "THERE IS NO WARRANTY; USE THIS PRODUCT AT YOUR OWN RISK.";

	final static long serialVersionUID = 817911632652898426L;

	int capacity;

	Packet packet = null;

	// The packets currently in transit.
	Packet[] array;

	// Index into array where the next packet sent should go.
	int sendPtr = 0;

	// Index into array where the next packet received should come from.
	int receivePtr = 0;

	// Number of slots in array currently in use.
	int usedSlots = 0;

	// Number of senders who have called setSender() but are not closed.
	int senderCount = 0;

	// The unique receiver. We need to activate() it whenever
	// packets arrive, or if connection is closed and receiver is MustRun.
	Component receiver = null;

	// The component who happens to be sending on a send()
	Component sender;

	LinkedList<Component> senders = null;

	// The list of types (Class objects) that senders have declared.
	// Vector senderTypes;
	// This network
	Network network; // added for subnet support

	// The type that the receiver has declared.
	Class receiverType;

	/**
	 * Constructor: make a new connection of a given size
	 */
	Connection(int size) {
		array = new Packet[size];
	}

	/**
	 * Invoked to tell us we have a(nother) sender.
	 */
	synchronized void bumpSenderCount(OutputPort op) {
		senderCount++;

	}

	/**
	 * Returns the number of packets this Connection can hold.
	 */
	public int capacity() {
		return capacity;
	}

	/**
	 * count = size
	 */
	public synchronized int count() {
		return usedSlots;
	}

	/**
	 * Throws an error if the input object types and output object types do not
	 * match (This needs some more thinking about...)
	 */
	public void checkTypes() {

		return;
		/*
		 * if (senderTypes == null) return; if (receiverType == Object.class)
		 * return; Enumeration types = senderTypes.elements(); while
		 * (types.hasMoreElements()) { Class senderType =
		 * (Class)(types.nextElement()); if
		 * (!(receiverType.isAssignableFrom(senderType)))
		 * FlowError.complain("Connection type mismatch"); }
		 * 
		 */

	}

	/**
	 * The close input connection function - this will prevent any more packets
	 * from being received from this connection. Need to interrupt sender(s).
	 */

	public synchronized void close() {

		network.GenTraceLine("Close input connection: " + receiver.getName());

		if (isClosed())
			return;
		senderCount = 0; // set sender count to zero
		if (usedSlots > 0)
			network.GenTraceLine(usedSlots
					+ " packets on input connection lost");

		notifyAll(); // wakes up any senders waiting for slots

	}

	/**
	 * Close the sending Component -- may have to be changed to specify
	 * component
	 */
	public synchronized void closeSender() {

		if (!isClosed()) {
			--senderCount;

			if (isDrained()) { // means closed AND empty
				// closed means senderCount == 0
				/* if (!(receiver instanceof MustRun) ) */
				receiver.terminating = true;

			}
			if (receiver.terminating) {
				if (receiver.getStatus() == StatusValues.DORMANT
						|| receiver.getStatus() == StatusValues.NOT_STARTED)
					receiver.activate();
				else
					notify(); // notify receiver
				//receiver.activate();
				//notify();

			}
		}
	}

	/**
	 * Invoked to get receiver.
	 */

	public Component getReceiver() { // added for subnet support
		return receiver;
	}

	/**
	 * Returns true if this connection is closed (not necessarily drained).
	 */

	public synchronized boolean isClosed() {
		return (senderCount == 0);
	}

	/**
	 * Returns true if this connection is drained (closed and empty).
	 */

	public synchronized boolean isDrained() {
		return (isClosed() && isEmpty());
	}

	/**
	 * Returns true if this connection is empty
	 */

	public synchronized boolean isEmpty() {
		return (usedSlots == 0);
	}

	/**
	 * Returns true if this connection is full.
	 */

	public synchronized boolean isFull() {
		return (usedSlots == array.length);
	}

	/**
	 * The receive function. See InputPort.receive.
	 */

	public synchronized Packet receive() {

		network.GenTraceLine("Receiving: " + receiver.getName());

		if (isDrained()) {
			network.GenTraceLine("Recv/close: " + receiver.getName());
			return null;
		}

		if (isEmpty()) {

			receiver.status = StatusValues.SUSP_RECV;
			network.GenTraceLine("Recv/susp: " + receiver.getName());

			try {
				wait();
			} catch (InterruptedException e) {
				throw new ThreadDeath();
			}
			if (isDrained()) {
				network.GenTraceLine("Receive drained: " + receiver.getName());
				return null;
			}
			receiver.status = StatusValues.ACTIVE;
			network.GenTraceLine("Recv/resume: " + receiver.getName());

		}

		if (isDrained()) {
			network.GenTraceLine("Receive drained: " + receiver.getName());
			return null;
		}
		Packet packet = array[receivePtr];
		array[receivePtr] = null;
		receivePtr = (receivePtr + 1) % array.length;
		notifyAll(); // notify any components waiting to send
		usedSlots--;

		packet.setOwner(receiver);
		String oc[] = { "open", "close" };
		String msg = "Receive OK: " + receiver.getName();
		if (Packet.NORMAL != packet.getType())
			msg += "; " + oc[packet.getType() - 1];
		if (null != packet.getContent())
			msg += "; " + packet.getContent();
		network.GenTraceLine(msg);

		network.active = true;
		return packet;
	}

	/**
	 * The send function. See OutputPort.send.
	 */

	synchronized boolean send(Packet packet, OutputPort op) {
		sender = op.sender;
		String msg = "Sending: " + op.sender.getName();
		String oc[] = { "open", "close" };
		if (Packet.NORMAL != packet.getType())
			msg += "; " + oc[packet.getType() - 1];
		if (null != packet.getContent())
			msg += "; " + packet.getContent();
		network.GenTraceLine(msg);
		if (isClosed()) {
			network.GenTraceLine("Send/close: " + op.sender.getName());
			return false;
		}

		// if (isEmpty())
		// receiver.activate();

		while (isFull()) {
			sender.status = StatusValues.SUSP_SEND;
			network.GenTraceLine("Send/susp: " + sender.getName());
			try {
				wait();
			} catch (InterruptedException e) {
				throw new ThreadDeath();
			}

			sender = op.sender;
			sender.status = StatusValues.ACTIVE;
			network.GenTraceLine("Send/resume: " + sender.getName());

		}

		if (isClosed()) {
			network.GenTraceLine("Send/close: " + op.sender.getName());
			return false;
		}
		packet.clearOwner();
		array[sendPtr] = packet;
		sendPtr = (sendPtr + 1) % array.length;
		if (receiver.getStatus() == StatusValues.DORMANT
				|| receiver.getStatus() == StatusValues.NOT_STARTED)
			receiver.activate();
		else
			notifyAll(); // notify receiver
		// other components waiting to send to this connection may also get
		// notified,
		// but this is handled by while statement above
		usedSlots++;

		op.sender.status = StatusValues.ACTIVE;
		// notifyAll(); if was empty, notify receiver
		// other components waiting to send to this connection may also get
		// notified,
		// but this is handled by while statement above

		network.GenTraceLine("Send OK: " + op.sender.getName());
		sender = null;
		network.active = true;
		return true;
	}

	/**
	 * Invoked to tell us we have a receiver.
	 */

	public void setReceiver(Component newReceiver) { // added for subnet
														// support
		receiver = newReceiver;
	}

	/**
	 * Invoked to tell us the type of packet content being sent or expected. The
	 * receiver's type must be a supertype of every sender's type, or the
	 * network is ill-formed.
	 */

	public void setType(Class type) {

		if (type == null)
			return;

		receiverType = type;

	}
}
