package com.jpmorrsn.javaFBP;

import java.util.*;
import java.util.concurrent.*;
import java.io.*;

/**
 * The abstract class which all flow networks extend directly or indirectly. A
 * specific flow network must override the <code>define()</code> method, which
 * is written using the <i>mini-language</i> (actually just highly restricted
 * Java invoking the <code>protected</code> methods of this class). The
 * mini-language specifies what threads are to be created using which
 * components, and what connections are established between the ports of those
 * components. Here is a sample:
 * 
 * <pre>
 * //		connect(component(&quot;Read&quot;, ReadText.class),
 * //			port(&quot;OUT&quot;),
 * //			component(&quot;Write&quot;, WriteText.class),
 * //			port(&quot;IN&quot;));
 * //		initialize(new InputStreamReader(System.in),
 * //			component(&quot;Read&quot;),
 * //			port(&quot;SOURCE&quot;));
 * //		initialize(new OutputStreamWriter(System.out),
 * //			component(&quot;Write&quot;),
 * //			port(&quot;DESTINATION&quot;));
 * </pre>
 * 
 * The effect of the above is to create two threads named <code>Read</code>
 * (an instance of <code>ReadText</code>) and <code>Write</code> (an
 * instance of <code>WriteText</code>, connect the <code>OUT</code> port of
 * the former to the <code>IN</code> port of the latter, and initialize the
 * <code>SOURCE</code> and <code>DESTINATION</code> parameter ports with
 * suitable Java objects for reading the standard input and writing the standard
 * output.
 */

public abstract class Network extends Component {

	static final String copyright = "Copyright 1999, 2000, 2001, 2002, J. Paul Morrison.  At your option, you may copy, "
			+ "distribute, or make derivative works under the terms of the Clarified Artistic License, "
			+ "based on the Everything Development Company's Artistic License.  A document describing "
			+ "this License may be found at http://www.jpaulmorrison.com/fbp/artistic2.htm. "
			+ "THERE IS NO WARRANTY; USE THIS PRODUCT AT YOUR OWN RISK.";

	protected static int DEBUGSIZE = 1;

	protected static int PRODUCTIONSIZE = 10;

	// int defaultCapacity = DEBUGSIZE; // change this when you go to production
	static int defaultCapacity = PRODUCTIONSIZE; // use this one for

	// production

	boolean debugging = false; // set on in network definition - local to

	// network/subnet

	boolean active = false; // used for deadlock detection - changed to non

	// static

	HashMap<String, Component> components = new HashMap<String, Component>();

	CountDownLatch cdl = null;

	PrintStream ps = null; // for debugging

	InputPort inport;

	OutputPort outport;

	static Network thisNetwork;

	// InputPort substreamSensitive;

	// OutputPort netsrc;

	// InputPort netdst;

	// static Thread outputCopier;

	/**
	 * Drive define method of network
	 */

	void callDefine() {

		try {
			define();
		} catch (Throwable t) {
			FlowError.complain(t.getMessage());
		}
		;
	}

	/**
	 * Returns a Component class object, given the component name in the network
	 */

	protected final Component component(String name) {
		Component comp = (components.get(name));
		if (comp == null)
			FlowError.complain("Reference to unknown component" + name);
		return comp;
	}

	/**
	 * Stores the Component class object with its network name in the Hashtable
	 * called 'components'
	 */

	protected final Component component(String name, Class type) {

		if (components.get(name) != null)
			FlowError.complain("Attempt to redefine component " + name);
		Component comp = null;
		try {
			comp = (Component) (type.newInstance());
		} catch (IllegalAccessException e) {
			FlowError.complain("Illegal access to component");
			return null; // unreachable
		} catch (InstantiationException e) {
			FlowError.complain("Cannot instantiate component");
			return null; // unreachable
		}
		;
		comp.setName(name);
		components.put(name, comp);
		comp.network = getNetwork();
		comp.mother = this;
		comp.status = StatusValues.NOT_STARTED;

		return comp;
	}

	/**
	 * Connect an output port of one component to an input port of another,
	 * using default capacity
	 */
	protected final void connect(Component sender, String outName,
			Component receiver, String inName) {

		connect(sender, outName, receiver, inName, 0);

	}

	/**
	 * Connect an output port of one component to an input port of another,
	 * specifying a connection capacity
	 */

	protected final void connect(Component sender, String outName,
			Component receiver, String inName, int size) {
		int cap = size;
		if (outName.equals("*"))
			outName = "*OUT";
		if (inName.equals("*"))
			inName = "*IN";
		if (sender.outputPorts.get(outName) != null)
			FlowError.complain("Multiple connections to output port:"
					+ sender.getName() + ' ' + outName);
		Object o = receiver.inputPorts.get(inName);
		if (o == null) {
			if (size == 0)
				cap = defaultCapacity;
			Connection c = new Connection(cap);
			c.setReceiver(receiver);
			c.network = getNetwork(); // added for tracing
			receiver.inputPorts.put(inName, c);
			o = c;
		}
		if (o instanceof Connection) {
			if (size != 0 && size != cap)
				FlowError
						.complain("Connection capacity does not agree with previous specification");
			OutputPort op = new OutputPort();
			Connection c = (Connection) (o);
			op.cnxt = c;
			op.setSender(sender);
			c.bumpSenderCount(op);
			sender.outputPorts.put(outName, op);
			if (c.senders == null) {
				c.senders = new LinkedList<Component>();
			}
			c.senders.add(sender);

		} else
			FlowError.complain("Mixed connection to input port " + inName);
	}

	protected abstract void define() throws Throwable;

	public Iterator enumerateComponents() {
		return (components.values()).iterator();
	}

	/**
	 * Execute method used by network being used as if it were a component
	 * 
	 * protected final void execute() { Packet ss =
	 * substreamSensitive.receive(); if (ss == null) { callDefine(); initiate();
	 * copyToOutput(); Packet p; while ((p = inport.receive()) != null) { if
	 * (!(netsrc.send(p))) { p.drop(); break; } } waitForAll(); } else {
	 * ss.drop(); int depth = 0; Packet p; while ((p = inport.receive()) !=
	 * null) { if (p.type == Packet.OPEN) depth++; else if (p.type ==
	 * Packet.CLOSE) depth--; if (depth == 0) { callDefine(); initiate();
	 * copyToOutput(); } if (!(netsrc.send(p))) { p.drop(); break; } if (depth ==
	 * 0) { netsrc.close(); waitForAll(); } } } }
	 */
	public void execute() {
	}

	/**
	 * Generate a trace line on the JAVAFLOW log node
	 */
	public void GenMonitorLine(String s) {

		if (debugging) {
			System.out.println("Monitor: " + s);
			if (System.out.checkError()) {
				FlowError.complain("Trace error");

			}
		}
	}

	/**
	 * Generate a trace line on the JAVAFLOW log node
	 */
	public void GenTraceLine(String s) {

		if (debugging) {
			// System.out.println( "Network: " + s );
			ps.println("Network: " + s);

			if (ps.checkError()) {
				FlowError.complain("Trace error");

			}
		}

	}

	/**
	 * Return network object
	 * 
	 * @return com.jpmorrsn.javaFBP.Network
	 */
	public static Network getNetwork() {
		return thisNetwork;
	}

	/**
	 * Execute network as a whole
	 */
	public final void go() {

		long now = System.currentTimeMillis();

		boolean verbose;
		verbose = System.getProperty("javaFBP.verbose", "false")
				.equalsIgnoreCase("true");

		if (verbose)
			debugging = true;

		thisNetwork = this;

		try {
			ps = new PrintStream(new FileOutputStream(
					"com\\jpmorrsn\\javaFBP\\test\\data\\trace"));
		} catch (FileNotFoundException e) {
			System.out.println("File not found");
		}

		try {

			callDefine();

			GenTraceLine("Run started");
			active = true;
			initiate();
			// activateAll();
			if (!debugging) {
				DeadlockDetect p = new DeadlockDetect(20, this);
				// deadlock detected after 20-40 secs. inactivity
				p.start();
				waitForAll();
				p.interrupt();
			} else {
				waitForAll();
			}
		} catch (FlowError e) {
			String s = "Flow Error :" + e;
			System.out.println("Network: " + s);
		}

		if (debugging) {
			long duration = System.currentTimeMillis() - now;
			long s = (long) duration / 1000;
			long ms = (long) duration % 1000;
			String mss = "000";
			String ms2 = mss.concat(String.valueOf(ms));
			int i = ms2.length();
			String ms3 = ms2.substring(i - 3, i);
			GenTraceLine("Run complete.  Time: " + s + '.' + ms3 + " seconds");

		}

	}

	void indicateTerminated(Component comp) {
		synchronized (comp) {
			comp.status = StatusValues.TERMINATED;
		}
		GenTraceLine("Terminated: " + comp.getName());

		cdl.countDown();
		// net.interrupt();
	}

	/**
	 * Build InitializationConnection object
	 */

	protected final void initialize(Object content, Component receiver,
			String inName) {
		// if (inName.equals("*")) inName = "*IN";
		if (receiver.inputPorts.get(inName) != null)
			FlowError.complain("Mixed connection to input port" + inName);
		InitializationConnection ic = new InitializationConnection(content,
				receiver);

		receiver.inputPorts.put(inName, ic);
	}

	/**
	 * Go through components opening ports, and activating those which are
	 * self-starting (have no input connections)
	 */
	void initiate() {
		cdl = new CountDownLatch(components.size());

		// go through all components twice!
		Iterator all = (components.values()).iterator();
		while (all.hasNext()) {
			Component comp = (Component) (all.next());
			comp.openPorts();
		}

		all = (components.values()).iterator();
		while (all.hasNext()) {
			Component comp = (Component) (all.next());

			comp.selfStarting = true;

			Iterator allin = (comp.inputPorts.values()).iterator();
			while (allin.hasNext()) {

				InputPort port = (InputPort) allin.next();
				if (port instanceof Connection) {
					Connection cnxt = (Connection) port;
					cnxt.checkTypes();
					comp.selfStarting = false;
					break;
				}

			}

			if (comp.selfStarting)
				comp.activate();
		}
	}

	/**
	 * Interrupt all components - used by DeadlockDetect
	 */

	public void interruptAll() {
		Iterator all = (components.values()).iterator();
		while (all.hasNext()) {
			Component comp = (Component) (all.next());
			comp.interrupt();
		}

	}

	protected Object[] introspect() {
		return new Object[] {
				"operates a general network defining a structure of "
						+ "components connected by their ports", "IN", "input",
				Object.class, "general input to the whole network", "OUT",
				"output", Object.class,
				"general output from the whole network", "SUBSTRING-SENSITIVE",
				"parameter", null, "makes input substring-sensitive if present" };
	}

	/**
	 * method establishing input connection for subnet
	 */
	/*
	 * protected final void networkInput(Component receiver, String inName) {
	 * connect(this, "*NETSOURCE", receiver, inName); netsrc =
	 * openOutput("*NETSOURCE"); }
	 */
	/**
	 * method establishing output connection for subnet
	 */
	/*
	 * protected final void networkOutput(Component sender, String outName) {
	 * connect(sender, outName, this, "*NETSINK"); netdst =
	 * openInput("*NETSINK"); }
	 */
	/**
	 * method to open ports for subnet
	 */

	protected void openPorts() { // dropped final - should never get executed
		inport = openInput("?");
		outport = openOutput("?");
		// substreamSensitive = openInput("SUBSTREAM-SENSITIVE");
	}

	/**
	 * method to register a port name for subnet
	 */

	protected final String port(String name) {
		int m;
		if (name.equals("*"))
			return name;
		else if (name.indexOf('*') != -1) {
			FlowError.complain("Stray * in port name");
			return name; // unreachable
		} else if ((m = name.indexOf('[')) != -1) {

			int n = name.indexOf("]", m);
			if (n == -1) {
				FlowError.complain("Port array index error");
				return null;
			} else {
				String x = name.substring(m + 1, n);
				Integer ix = new Integer(x);
				return port(name.substring(0, m), ix.intValue());
			}
		} else
			return name;
	}

	/**
	 * method to register an array port with index
	 */

	protected final String port(String name, int index) {
		return "*" + name + "*" + (char) index;
	}

	/**
	 * Test if network as a whole has terminated
	 */

	void waitForAll() {
		long i = cdl.getCount();
		i = i + 0;
		try {
			cdl.await();
		} catch (InterruptedException e) {
			FlowError.complain("Component " + getName() + " interrupted");
		}

		GenTraceLine(getName() + " terminating.");
	}

}
