package com.jpmorrsn.javaFBP;

import java.util.*;



/** A Packet may either contain an Object, when <code> type</type> is <code> NORMAL </code>,
 * or a String, when <code>type</code> is not <code>NORMAL</code>.  The latter case
 * is used for things like open and close brackets (where the String will be the name
 * of a group. e.g. accounts)
*/

public class Packet {

static final String copyright = 
"Copyright 1999, 2000, 2001, 2002, J. Paul Morrison.  At your option, you may copy, " +
"distribute, or make derivative works under the terms of the Clarified Artistic License, " +
"based on the Everything Development Company's Artistic License.  A document describing " +
"this License may be found at http://www.jpaulmorrison.com/fbp/artistic2.htm. " +
"THERE IS NO WARRANTY; USE THIS PRODUCT AT YOUR OWN RISK.";

	public static final int OPEN = 1;
	public static final int CLOSE = 2;

	public static final int NORMAL = 0;


	Object content;

	int type;

	Object owner;

	HashMap<String,Chain> chains = null;

	HashMap<String,Object> attrs = null;

	// An iteration that has nothing to iterate.               
	static Iterator nullIter = (new HashMap().values()).iterator();  

	long packetNumber = 0;  // for debugging only


	public Packet(int newType, String newName, Thread newOwner) {
		content = newName;
		setOwner(newOwner);
		type = newType;
		}
	public Packet(Object newContent, Thread newOwner) {
		content = newContent;
		setOwner(newOwner);
		type = NORMAL;
		}
	/** Maintains a Chain of Packets attached to this Packet.
	* A Packet may have multiple named Chains attached to it, accessed via a HashMap.
	* Since Packets are attached to Chains, as well as Chains to Packets,
	* this results in an alternation of Packets and Chains, creating a tree structure.
	*/

	public void attach(String name, Packet subordinate) {
		if (chains == null)
			chains = new HashMap<String,Chain>();                     
		Chain chain = chains.get(name);
		if (chain == null) {
			chain = new Chain();
			chain.head = this;
			chain.members = new HashSet<Packet>();
			chain.name = name;
			chains.put(name, chain);
			}
		Packet root = getRoot();
		if (subordinate.owner != root.owner)
			FlowError.complain("Attaching unowned packet");
		if (subordinate == root)
			FlowError.complain("Loop in packet chain");
		subordinate.setOwner(chain);
		chain.members.add(subordinate);
		}
	/** Clear the owner of a Packet, and reduce the number of Packets owned by the owner
	*  (if owner is a Component) - if Packet is chained, owner is Chain
	*/

	void clearOwner() {
		if (owner instanceof Component) {
			Component c = (Component)(owner);
			c.packetCount--;

	
			}
		owner = null;
		}
	/** Detach Packet from named chain
	*/

	public boolean detach(String name, Packet subordinate) {
		if (chains == null) return false;
		Chain chain = chains.get(name);
		if (chain == null) return false;
		if (!(chain.members.remove(subordinate))) return false;
		Packet root = getRoot();
		subordinate.setOwner(root.owner);
		return true;
		}
/** Drop Packet and clear owner reference.
	* Note: Java will not reclaim the space until after all references
	*  to this packet have been cleared - ideally, one should not use this
	* knowledge, but bright people probably will!
	*/
	public void drop() {

		// this test seems a bit flaky...
		Thread t = Thread.currentThread();
		if (t != owner)
		   FlowError.complain("Packet not owned by current component, or component has terminated");

		clearOwner();
		}
	/** Get named attribute of Packet - may be any Object
	*/
	public Object getAttribute(String key) {
		if (attrs != null)
			return attrs.get(key);
		else
			return null;
		}
	/** Get all attributes of this Packet (as Iterator)
	*/
	public Iterator getAttributes() {
		if (attrs != null)
			return (attrs.keySet()).iterator();
		else
			return nullIter;
		}
	/** Get named chain (as Iterator)
	*/

	public Iterator getChain(String name) {
		if (chains == null)
			return nullIter;
		Chain chain = chains.get(name);
		if (chain != null)
			return chain.members.iterator();        
		else
			return nullIter;
		}
	/** Get all chains for this Packet (as Iterator)
	*/

	public Iterator getChains() {
		if (chains != null)
			return (chains.keySet()).iterator();
		else
			return nullIter;
		}
	/** Get contents of this Packet - may be any Object; if <code>type</code> is not
	* <code>NORMAL</code> it returns <code>null</code>
	*/

	public Object getContent() {
		//if (type == NORMAL)
			return content;
		// else
			// return null;
		}
	/** Get contents of this Packet for case where <code>type</code> is not
	* <code>NORMAL</code>; if it is <code>NORMAL</code>, it returns <code>null</code>
	*/
	   public String getName() {
		if (type == NORMAL)
			return null;
		else
			return (String)(content);
		}
	/** Get root of this Packet - it follows the Packet owner chain up until
	* it finds a Packet that is owned by a Component rather than by a Chain
	*/

	Packet getRoot() {
		Packet p = this;
		while (p.owner instanceof Chain)
			p = ((Chain)(p.owner)).head;
		return p;
		}
	/** This method returns the type of a Packet
	*/

	public int getType() {
		return type;
		}
	/** Make an Object a named attribute of a Packet
	*/

	 public void putAttribute(String key, Object value) {
		 
		 if (attrs == null)
			attrs = new HashMap<String, Object>();                   
		 attrs.put(key, value);
		}
	/** Remove a named attribute from a Packet (does not return the attribute)
	*/

	public void removeAttribute(String key) {
		if (attrs != null)
			attrs.remove(key);
		}
	/** Change the owner of a Packet - if the owner is a Component,
	* increment the number of Packets owned by that Component
	* (when the Component is deactivated, it must no longer own any Packets)
	*/

	void setOwner(Object newOwner) {
		clearOwner();
		owner = newOwner;
		if (owner instanceof Component) {
			Component c = (Component)(owner);
			c.packetCount++;  // count of owned packets

	
			}
		}
	
}
